from __future__ import absolute_import, division, print_function
import json
import numpy as np
import os
import six

import bcolz
from pybedtools import BedTool
import pyBigWig
from pysam import FastaFile

from .util import makedirs
from .util import one_hot_encode_sequence
from .util import nan_to_zero

NUM_SEQ_CHARS = 4

_blosc_params = bcolz.cparams(clevel=5, shuffle=bcolz.SHUFFLE, cname='lz4')

_array_writer = {
    'numpy': lambda arr, path: np.save(path, arr),
    'bcolz': lambda arr, path: bcolz.carray(
        arr, rootdir=path, cparams=_blosc_params, mode='w').flush()
}


def extract_fasta_to_file(fasta, output_dir, mode='bcolz', overwrite=False):
    assert mode in _array_writer

    makedirs(output_dir, exist_ok=overwrite)
    fasta_file = FastaFile(fasta)
    file_shapes = {}
    for chrom, size in zip(fasta_file.references, fasta_file.lengths):
        data = np.zeros((size, NUM_SEQ_CHARS), dtype=np.float32)
        seq = fasta_file.fetch(chrom)
        one_hot_encode_sequence(seq, data)
        file_shapes[chrom] = data.shape
        _array_writer[mode](data, os.path.join(output_dir, chrom))

    with open(os.path.join(output_dir, 'metadata.json'), 'w') as fp:
        json.dump({'file_shapes': file_shapes,
                   'type': 'array_{}'.format(mode),
                   'source': fasta}, fp)


def extract_bigwig_to_file(bigwig, output_dir, mode='bcolz', dtype=np.float32,
                           overwrite=False, nan_as_zero=True):
    assert mode in _array_writer

    makedirs(output_dir, exist_ok=overwrite)
    bw = pyBigWig.open(bigwig)
    chrom_sizes = bw.chroms()
    file_shapes = {}
    for chrom, size in six.iteritems(chrom_sizes):
        data = np.zeros(size, dtype=np.float32)
        data[:] = bw.values(chrom, 0, size)
        if nan_as_zero:
            nan_to_zero(data)
        _array_writer[mode](data.astype(dtype),
                            os.path.join(output_dir, chrom))
        file_shapes[chrom] = data.shape
    bw.close()

    with open(os.path.join(output_dir, 'metadata.json'), 'w') as fp:
        json.dump({'file_shapes': file_shapes,
                   'type': 'array_{}'.format(mode),
                   'source': bigwig}, fp)


def read_genome_sizes(genome_file):
    with open(genome_file) as fp:
        chr2size = {}
        for line in fp:
            chrom, size = line.split()
            chr2size[chrom] = int(size)
    return chr2size


def load_directory(base_dir, in_memory=False):
    with open(os.path.join(base_dir, 'metadata.json'), 'r') as fp:
        metadata = json.load(fp)

    if metadata['type'] == 'array_numpy':
        mmap_mode = None if in_memory else 'r'
        data = {chrom: np.load('{}.npy'.format(os.path.join(base_dir, chrom)),
                                mmap_mode=mmap_mode)
                for chrom in metadata['file_shapes']}

        for chrom, shape in six.iteritems(metadata['file_shapes']):
            if data[chrom].shape != tuple(shape):
                raise ValueError('Inconsistent shape found in metadata file: '
                                 '{} - {} vs {}'.format(chrom, shape,
                                                        data[chrom].shape))
    elif metadata['type'] == 'array_bcolz':
        data = {chrom: bcolz.open(os.path.join(base_dir, chrom), mode='r')
                for chrom in metadata['file_shapes']}
        if in_memory:
            data = {k: data[k].copy() for k in data.keys()}

        for chrom, shape in six.iteritems(metadata['file_shapes']):
            if data[chrom].shape != tuple(shape):
                raise ValueError('Inconsistent shape found in metadata file: '
                                 '{} - {} vs {}'.format(chrom, shape,
                                                        data[chrom].shape))
    else:
        raise ValueError('Can only extract from array_bcolz and array_numpy')

    return data



from __future__ import absolute_import

# Setup logging
import logging

log_formatter = \
    logging.Formatter('%(levelname)s:%(asctime)s:%(name)s] %(message)s')
_logger = logging.getLogger('genomelake')
_handler = logging.StreamHandler()
_handler.setLevel(logging.DEBUG)
_handler.setFormatter(log_formatter)
_logger.setLevel(logging.DEBUG)
_logger.addHandler(_handler)



import os
from setuptools.extension import Extension
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='genomelake',

    version='0.1.4',

    description='Simple and efficient random access to genomic data for deep learning models.',
    long_description=read('README.md'),

    url='https://github.com/kundajelab/genomelake',

    author='Chuan-Sheng Foo',
    author_email='csfoo@cs.stanford.edu',

    license='BSD-3',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: Linux',
    ],

    keywords='deeplearning neuralnets genomics',

    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    setup_requires=['cython'],

    ext_modules=[Extension('genomelake.util', ['genomelake/util.pyx'])],

    install_requires=['bcolz>=1.1', 'numpy', 'pybedtools',
                      'pyBigWig', 'pysam', 'six>=1.9.0'],
    extras_require={"develop": ['coveralls', 'pytest>=3.3.1', 'pytest-cov']}
)



# cython: embedsignature=True
import os
import sys

from libc.math cimport isnan


def makedirs(path, mode=0777, exist_ok=False):
    try:
        os.makedirs(path, mode)
    except OSError:
        if not exist_ok or not os.path.isdir(path):
            raise
        else:
            sys.stderr.write('Warning: directory %s exists.\n' % path)


cdef inline int char2index(char ch) except -2:
    if ch == 'A' or ch == 'a':
        return 0
    if ch == 'C' or ch == 'c':
        return 1
    if ch == 'G' or ch == 'g':
        return 2
    if ch == 'T' or ch == 't':
        return 3
    if ch == 'N' or ch == 'n':
        return -1
    raise ValueError('Invalid base encountered.')


cpdef void one_hot_encode_sequence(str seq, float[:, :] encoded) except *:
    cdef size_t row_idx = 0
    cdef int col_idx

    if encoded.shape[0] != len(seq):
        raise ValueError('encoded array not the same length as given seq')

    if encoded.shape[1] != 4:
        raise ValueError('encoded array needs to have 4 columns')
    
    for base in seq:
        col_idx = char2index(ord(base))

        if col_idx >= 0:
            encoded[row_idx, col_idx] = 1
        else:
            encoded[row_idx, :] = 0.25

        row_idx += 1


cpdef nan_to_zero(float[:] arr):
    cdef Py_ssize_t k
    for k in range(arr.size):
        if isnan(arr[k]):
            arr[k] = 0



from __future__ import absolute_import, division, print_function
import numpy as np
import six

import bcolz
from pybedtools import BedTool
from pybedtools import Interval
from pysam import FastaFile
import pyBigWig

from . import backend
from .util import nan_to_zero
from .util import one_hot_encode_sequence

NUM_SEQ_CHARS = 4


class BaseExtractor(object):
    dtype = np.float32

    def __init__(self, datafile, **kwargs):
        self._datafile = datafile

    def __call__(self, intervals, out=None, **kwargs):
        data = self._check_or_create_output_array(intervals, out)
        self._extract(intervals, data, **kwargs)
        return data

    def _check_or_create_output_array(self, intervals, out):
        width = intervals[0].stop - intervals[0].start
        output_shape = self._get_output_shape(len(intervals), width)

        if out is None:
            out = np.zeros(output_shape, dtype=self.dtype)
        else:
            if out.shape != output_shape:
                raise ValueError('out array has incorrect shape: {} '
                                 '(need {})'.format(out.shape, output_shape))
            if out.dtype != self.dtype:
                raise ValueError('out array has incorrect dtype: {} '
                                 '(need {})'.format(out.dtype, self.dtype))
        return out

    def _extract(self, intervals, out, **kwargs):
        'Subclassses should implement this and return the data'
        raise NotImplementedError

    @staticmethod
    def _get_output_shape(num_intervals, width):
        'Subclasses should implement this and return the shape of output'
        raise NotImplementedError


class ArrayExtractor(BaseExtractor):

    def __init__(self, datafile, in_memory=False, **kwargs):
        super(ArrayExtractor, self).__init__(datafile, **kwargs)
        self._data = backend.load_directory(datafile, in_memory=in_memory)
        self.multiprocessing_safe = in_memory

        arr = next(iter(self._data.values()))
        def _mm_extract(self, intervals, out, **kwargs):
            mm_data = self._data
            for index, interval in enumerate(intervals):
                out[index] = mm_data[interval.chrom][interval.start:interval.stop]

        # output shape method
        shape = arr.shape
        if len(shape) == 1:
            def _get_output_shape(num_intervals, width):
                return (num_intervals, width)
        elif len(shape) == 2:
            def _get_output_shape(num_intervals, width):
                return (num_intervals, width, shape[1])
        else:
            raise ValueError('Can only extract from 1D/2D arrays')

        self._mm_extract = _mm_extract.__get__(self)
        self._extract = self._mm_extract
        self._get_output_shape = staticmethod(_get_output_shape).__get__(self)


class FastaExtractor(BaseExtractor):

    def __init__(self, datafile, use_strand=False, **kwargs):
        super(FastaExtractor, self).__init__(datafile, **kwargs)
        self.use_strand = use_strand
        self.fasta = FastaFile(self._datafile)

    def _extract(self, intervals, out, **kwargs):    
        for index, interval in enumerate(intervals):
            seq = self.fasta.fetch(str(interval.chrom), interval.start,
                                       interval.stop)
            one_hot_encode_sequence(seq, out[index, :, :])

            # reverse-complement seq the negative strand
            if self.use_strand and interval.strand == "-":
                out[index, :, :] = out[index, ::-1, ::-1]

        return out

    @staticmethod
    def _get_output_shape(num_intervals, width):
        return (num_intervals, width, NUM_SEQ_CHARS)


class BigwigExtractor(BaseExtractor):

    def __init__(self, datafile, **kwargs):
        super(BigwigExtractor, self).__init__(datafile, **kwargs)
        self._verbose = kwargs.get('verbose', False)

    def _extract(self, intervals, out, **kwargs):
        out[:] = self._bigwig_extractor(self._datafile, intervals,
                                        **kwargs)

        return out

    @staticmethod
    def _get_output_shape(num_intervals, width):
        return (num_intervals, width)

    @staticmethod
    def _bigwig_extractor(datafile, intervals, out=None, **kwargs):
        nan_as_zero = kwargs.get('nan_as_zero', True)
        if out is None:
            width = intervals[0].stop - intervals[0].start
            out = np.zeros((len(intervals), width), dtype=np.float32)

        bw = pyBigWig.open(datafile)
        for index, interval in enumerate(intervals):
            out[index] = bw.values(
                interval.chrom, interval.start, interval.stop)
            if nan_as_zero:
                nan_to_zero(out[index])
        bw.close()

        return out
